<?php

namespace Drupal\advanced_file_destination\Entity;

use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Entity\RevisionableInterface;
use Drupal\user\UserInterface;
use phpDocumentor\Reflection\DocBlock\Tags\Var_;

/**
 * Defines the Advanced File Destination Directory entity.
 *
 * @ConfigEntityType(
 *   id = "afd_directory",
 *   label = @Translation("File Destination Directory"),
 *   handlers = {
 *     "storage" = "Drupal\advanced_file_destination\AFDDirectoryStorage",
 *     "list_builder" = "Drupal\advanced_file_destination\Controller\DirectoryListBuilder",
 *     "form" = {
 *       "add" = "Drupal\advanced_file_destination\Form\AdvancedFileDestinationDirectoryForm",
 *       "edit" = "Drupal\advanced_file_destination\Form\AdvancedFileDestinationDirectoryForm",
 *       "delete" = "Drupal\advanced_file_destination\Form\AdvancedFileDestinationDeleteForm"
 *     },
 *     "route_provider" = {
 *       "html" = "Drupal\Core\Entity\Routing\AdminHtmlRouteProvider"
 *     }
 *   },
 *   config_prefix = "directory",
 *   admin_permission = "administer advanced file destination",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "name",
 *     "created" = "created",
 *     "changed" = "changed",
 *     "uuid" = "uuid",
 *     "status" = "status",
 *     "weight" = "weight",
 *     "revision" = "revision_id",
 *     "revision_log_message" = "revision_log"
 *   },
 *   revision_metadata_keys = {
 *     "revision_user" = "revision_uid",
 *     "revision_created" = "revision_timestamp",
 *     "revision_log_message" = "revision_log"
 *   },
 *   config_export = {
 *     "id",
 *     "name",
 *     "path",
 *     "status",
 *     "weight",
 *     "roles",
 *     "created",
 *     "changed",
 *     "revision_id",
 *     "revision_log"
 *   },
 *   links = {
 *     "add-form" = "/admin/content/directories/add",
 *     "canonical" = "/admin/content/directories/{afd_directory}/edit",
 *     "edit-form" = "/admin/content/directories/{afd_directory}/edit", 
 *     "delete-form" = "/admin/content/directories/{afd_directory}/delete",
 *     "collection" = "/admin/content/directories",
 *     "revision" = "/admin/content/directories/{afd_directory}/revisions"
 *   }
 * )
 */
class AdvancedFileDestinationDirectory extends ConfigEntityBase implements RevisionableInterface {

  /**
   * The directory ID.
   *
   * @var string
   */
  protected $id;

  /**
   * The directory name.
   *
   * @var string
   */
  protected $name;

  /**
   * The directory path.
   *
   * @var string
   */
  protected $path;

  /**
   * The directory status.
   *
   * @var bool
   */
  protected $status = TRUE;

  /**
   * The directory weight.
   *
   * @var int
   */
  protected $weight = 0;

  /**
   * The roles that can access this directory.
   *
   * @var array
   */
  protected $roles = [];

  /**
   * The revision ID.
   *
   * @var int
   */
  protected $revision_id;

  /**
   * The revision log message.
   *
   * @var string
   */
  protected $revision_log;

  /**
   * Indicates whether a new revision should be created.
   *
   * @var bool
   */
  protected $newRevision = FALSE;

  /**
   * The loaded revision ID.
   *
   * @var int
   */
  protected $loadedRevisionId;

  /**
   * Whether this is the default revision.
   *
   * @var bool
   */
  protected $isDefaultRevision = TRUE;

  /**
   * The creation timestamp.
   *
   * @var int
   */
  protected $created;

  /**
   * The change timestamp.
   *
   * @var int
   */
  protected $changed;

  /**
   * Gets the directory path.
   *
   * @return string
   *   The directory path.
   */
  public function getPath() {
    return $this->path;
  }

  /**
   * Sets the directory path.
   *
   * @param string $path
   *   The directory path.
   *
   * @return $this
   */
  public function setPath($path) {
    $this->path = $path;
    return $this;
  }

  /**
   * Gets the directory weight.
   *
   * @return int
   *   The directory weight.
   */
  public function getWeight() {
    return $this->weight;
  }

  /**
   * Sets the directory weight.
   *
   * @param int $weight
   *   The directory weight.
   *
   * @return $this
   */
  public function setWeight($weight) {
    $this->weight = $weight;
    return $this;
  }

  /**
   * Returns whether the directory is active.
   *
   * @return bool
   *   TRUE if the directory is active, FALSE otherwise.
   */
  public function isActive() {
    return $this->status;
  }

  /**
   * Sets the directory status.
   *
   * @param bool $status
   *   TRUE if the directory is active, FALSE otherwise.
   *
   * @return $this
   */
  public function setStatus($status) {
    $this->status = $status;
    return $this;
  }

  /**
   * Gets the roles that can access this directory.
   *
   * @return array
   *   The roles.
   */
  public function getRoles() {
    return $this->roles;
  }

  /**
   * Sets the roles that can access this directory.
   *
   * @param array $roles
   *   The roles.
   *
   * @return $this
   */
  public function setRoles(array $roles) {
    $this->roles = $roles;
    return $this;
  }

  /**
   * Gets the revision log message.
   *
   * @return string
   *   The revision log message.
   */
  public function getRevisionLogMessage() {
    return $this->revision_log;
  }

  /**
   * Sets the revision log message.
   *
   * @param string $revision_log
   *   The revision log message.
   *
   * @return $this
   */
  public function setRevisionLogMessage($revision_log) {
    $this->revision_log = $revision_log;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(EntityStorageInterface $storage, $update = TRUE) {
    parent::postSave($storage, $update);

    // Clear cached directories using the service container directly
    \Drupal::service('advanced_file_destination.manager')->clearCachedDirectories();
  }

  /**
   * {@inheritdoc}
   */
  public function label() {
    return $this->name;
  }

  /**
   * Gets the name.
   *
   * @return string
   *   The name.
   */
  public function getName() {
    return $this->name;
  }

  /**
   * Sets the name.
   *
   * @param string $name
   *   The name.
   *
   * @return $this
   */
  public function setName($name) {
    $this->name = $name;
    return $this;
  }

  /**
   * Gets the creation timestamp.
   *
   * @return int
   *   The creation timestamp or 0 for not yet saved directories.
   */
  public function getCreatedTime() {
    return $this->get('created')->value ?? \Drupal::time()->getRequestTime();
  }

  /**
   * Gets the change timestamp.
   *
   * @return int
   *   The changed timestamp or 0 for not yet saved directories.
   */
  public function getChangedTime() {
    return $this->get('changed')->value ?? 0;
  }

  /**
   * Gets the directory status.
   *
   * @return bool
   *   The directory status.
   */
  public function status() {
    if (is_bool($this->status)) {
      return $this->status;
    }
    return (bool) ($this->get('status')?->value ?? TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    parent::preSave($storage);

    // On creation, set created timestamp if not already set
    if ($this->isNew() && !isset($this->created)) {
      $this->created = \Drupal::time()->getRequestTime();
    }
    
    // Always set changed timestamp on save
    $this->changed = \Drupal::time()->getRequestTime();
  }

  /**
   * {@inheritdoc}
   */
  public function isNewRevision() {
    return $this->newRevision;
  }

  /**
   * {@inheritdoc}
   */
  public function setNewRevision($value = TRUE) {
    $this->newRevision = $value;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRevisionId() {
    return $this->revision_id;
  }

  /**
   * {@inheritdoc}
   */
  public function setRevisionId($revision_id) {
    $this->revision_id = $revision_id;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRevisionCreationTime() {
    return $this->revision_timestamp ?? NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function setRevisionCreationTime($timestamp) {
    $this->revision_timestamp = $timestamp;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRevisionUser() {
    if (isset($this->revision_uid)) {
      try {
        return \Drupal::entityTypeManager()->getStorage('user')->load($this->revision_uid);
      }
      catch (\Exception $e) {
        \Drupal::logger('advanced_file_destination')->error('Failed to load revision user: @error', ['@error' => $e->getMessage()]);
      }
    }
    return \Drupal::currentUser();
  }

  /**
   * {@inheritdoc}
   */
  public function setRevisionUser(UserInterface $account) {
    $this->set('revision_uid', $account->id());
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getRevisionUserId() {
    return $this->get('revision_uid') ? $this->get('revision_uid')->value : \Drupal::currentUser()->id();
  }

  /**
   * {@inheritdoc}
   */
  public function setRevisionUserId($user_id) {
    $this->revision_uid = $user_id;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getLoadedRevisionId() {
    return $this->loadedRevisionId;
  }

  /**
   * {@inheritdoc}
   */
  public function updateLoadedRevisionId() {
    $this->loadedRevisionId = $this->revision_id;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function isDefaultRevision($new_value = NULL) {
    if ($new_value !== NULL) {
      $this->isDefaultRevision = $new_value;
    }
    return $this->isDefaultRevision;
  }

  /**
   * {@inheritdoc}
   */
  public function wasDefaultRevision() {
    return isset($this->isDefaultRevision) ? $this->isDefaultRevision : TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function isLatestRevision() {
    if ($this->revision_id) {
      $entityTypeId = $this->getEntityTypeId();
      $revisionIds = \Drupal::entityTypeManager()
        ->getStorage($entityTypeId)
        ->getQuery()
        ->accessCheck(FALSE)
        ->allRevisions()
        ->condition($entityTypeId . '_id', $this->id())
        ->sort('revision_id', 'DESC')
        ->execute();
      if ($revisionIds) {
        $latest_revision_id = array_keys($revisionIds)[0];
        return $latest_revision_id == $this->getRevisionId();
      }
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function preSaveRevision(EntityStorageInterface $storage, \stdClass $record) {
    parent::preSave($storage);

    if (!$this->isNewRevision() && isset($this->original) && !$this->isLatestRevision()) {
      // If we are updating an existing revision, preserve the original author and
      // timestamp information.
      $record->revision_uid = $this->original->getRevisionUserId();
      $record->revision_timestamp = $this->original->getRevisionCreationTime();
    }
    else {
      $record->revision_uid = \Drupal::currentUser()->id();
      $record->revision_timestamp = \Drupal::time()->getRequestTime();
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type);

    $fields['name'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Name'))
      ->setDescription(t('The name of the directory.'))
      ->setRequired(TRUE)
      ->setSettings([
        'max_length' => 255,
        'text_processing' => 0,
      ])
      ->setDefaultValue('')
      ->setDisplayOptions('view', [
        'label' => 'visible',
        'type' => 'string',
        'weight' => -5,
      ])
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -5,
      ]);

    $fields['path'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Path'))
      ->setDescription(t('The directory path.'))
      ->setRequired(TRUE)
      ->setSettings([
        'max_length' => 255,
        'text_processing' => 0,
      ])
      ->setDefaultValue('')
      ->setDisplayOptions('view', [
        'label' => 'visible',
        'type' => 'string',
        'weight' => -4,
      ])
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -4,
      ]);

    $fields['status'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Status'))
      ->setDescription(t('Whether the directory is active.'))
      ->setDefaultValue(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'boolean_checkbox',
        'settings' => [
          'display_label' => TRUE,
        ],
        'weight' => -3,
      ]);

    $fields['weight'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Weight'))
      ->setDescription(t('The directory weight.'))
      ->setDefaultValue(0)
      ->setDisplayOptions('form', [
        'type' => 'number',
        'weight' => -2,
      ]);

    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Created'))
      ->setDescription(t('The time the directory was created.'))
      ->setDisplayOptions('view', [
        'label' => 'visible',
        'type' => 'timestamp',
        'weight' => -3,
      ]);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Updated'))
      ->setDescription(t('The time the directory was last updated.'))
      ->setDisplayOptions('view', [
        'label' => 'visible',
        'type' => 'timestamp',
        'weight' => -2,
      ]);

    $fields['revision_id'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Revision ID'))
      ->setDescription(t('The directory revision ID.'))
      ->setReadOnly(TRUE);

    $fields['revision_log'] = BaseFieldDefinition::create('string_long')
      ->setLabel(t('Revision log message'))
      ->setDescription(t('Briefly describe the changes you have made.'))
      ->setRevisionable(TRUE)
      ->setDefaultValue('')
      ->setDisplayOptions('form', [
        'type' => 'string_textarea',
        'weight' => 25,
      ]);

    $fields['revision_uid'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Revision user'))
      ->setDescription(t('The user ID of the author of the current revision.'))
      ->setSetting('target_type', 'user')
      ->setRevisionable(TRUE);

    $fields['revision_timestamp'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Revision timestamp'))
      ->setDescription(t('The time that the current revision was created.'))
      ->setRevisionable(TRUE);

    return $fields;
  }

}
